package ru.dbotthepony.mc.otm.android

import com.google.common.collect.ImmutableMap
import com.google.gson.GsonBuilder
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import net.minecraft.client.server.IntegratedServer
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.resources.ResourceManager
import net.minecraft.server.packs.resources.SimpleJsonResourceReloadListener
import net.minecraft.util.profiling.ProfilerFiller
import net.minecraftforge.common.MinecraftForge
import net.minecraftforge.event.AddReloadListenerEvent
import net.minecraftforge.event.OnDatapackSyncEvent
import net.minecraftforge.eventbus.api.Event
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.MINECRAFT_SERVER
import ru.dbotthepony.mc.otm.NULLABLE_MINECRAFT_SERVER
import ru.dbotthepony.mc.otm.SERVER_IS_LIVE
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.data.SerializedFunctionRegistry
import ru.dbotthepony.mc.otm.network.MatteryPacket
import ru.dbotthepony.mc.otm.network.RegistryNetworkChannel
import ru.dbotthepony.mc.otm.network.enqueueWork
import ru.dbotthepony.mc.otm.network.packetHandled
import ru.dbotthepony.mc.otm.onceServer
import java.util.LinkedList
import java.util.function.Supplier

typealias AndroidResultTransformer = SerializedFunctionRegistry.BoundFunction<Pair<AndroidResearch, AndroidFeature>, Unit>
typealias ComponentSupplier = SerializedFunctionRegistry.BoundFunction<Unit, Component>

object AndroidResearchManager : SimpleJsonResourceReloadListener(GsonBuilder().setPrettyPrinting().disableHtmlEscaping().create(), "otm_android_research"), Iterable<AndroidResearchType> {
	/**
	 * Feel free to register functions inside this thing from anywhere in your code
	 * (registration and querying is completely thread safe).
	 *
	 * Just make sure client and server has the same set of functions defined.
	 */
	val featureResultTransformers = SerializedFunctionRegistry<Pair<AndroidResearch, AndroidFeature>, Unit>()

	/**
	 * Feel free to register functions inside this thing from anywhere in your code
	 * (registration and querying is completely thread safe).
	 *
	 * Just make sure client and server has the same set of functions defined.
	 */
	val descriptionFuncs = SerializedFunctionRegistry<Unit, Component>()

	fun descriptionFunc(name: ResourceLocation, base: String, vararg argument: Supplier<Any>): ComponentSupplier {
		return descriptionFuncs.register(name) {->
			return@register TranslatableComponent(base, *argument.map { it.get() }.toTypedArray())
		}.bind()
	}

	private var firedRegistrationEvent = false

	/**
	 * Event-style registration of serializable functions, for those who prefer/need it
	 *
	 * Fired *once* on [MinecraftForge.EVENT_BUS] before loading android research
	 */
	object RegisterFuncsEvent : Event() {
		val manager get() = AndroidResearchManager
		val featureResults by ::featureResultTransformers
		val descriptionFunctions by ::descriptionFuncs

		fun descriptionFunc(name: ResourceLocation, base: String, vararg argument: Supplier<Any>): ComponentSupplier = AndroidResearchManager.descriptionFunc(name, base, *argument)
	}

	const val DIRECTORY = "otm_android_research"
	private val LOGGER = LogManager.getLogger()

	operator fun get(id: ResourceLocation): AndroidResearchType? {
		return researchMap[id] ?: maskedMap[id]
	}

	override fun iterator(): Iterator<AndroidResearchType> {
		return researchMap.values.iterator()
	}

	private val maskedMap = HashMap<ResourceLocation, AndroidResearchType>()

	/**
	 * Internal use only, used by [AndroidResearchDataProvider] for validation purposes
	 */
	internal fun put(value: AndroidResearchType) {
		maskedMap[value.id] = value
	}

	var researchMap: Map<ResourceLocation, AndroidResearchType> = mapOf()
		private set

	internal fun fireRegistrationEvent() {
		if (!firedRegistrationEvent) {
			try {
				MinecraftForge.EVENT_BUS.post(RegisterFuncsEvent)
			} finally {
				firedRegistrationEvent = true
			}
		}
	}

	override fun apply(
		jsonElementMap: Map<ResourceLocation, JsonElement>,
		manager: ResourceManager,
		profiler: ProfilerFiller
	) {
		fireRegistrationEvent()

		val builder = ImmutableMap.builder<ResourceLocation, AndroidResearchType>()

		for ((k, v) in jsonElementMap) {
			if (v !is JsonObject) {
				LOGGER.error("Skipping Android Feature $k because top value is not a Object")
				continue
			}

			builder.put(k, AndroidResearchType.fromJson(v, k))
		}

		researchMap = builder.build()

		for (value in researchMap.values) {
			value.validate()
		}

		if (SERVER_IS_LIVE) {
			onceServer {
				for (ply in MINECRAFT_SERVER.playerList.players) {
					ply.matteryPlayer?.reloadResearch()
				}
			}
		}
	}

	fun reloadEvent(event: AddReloadListenerEvent) {
		event.addListener(this)
	}

	fun syncEvent(event: OnDatapackSyncEvent) {
		val packet = SyncPacket(researchMap.values)

		if (event.player != null) {
			RegistryNetworkChannel.send(event.player!!, packet)
		} else {
			RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), packet)
		}
	}

	class SyncPacket(val collection: Collection<AndroidResearchType>) : MatteryPacket {
		override fun write(buff: FriendlyByteBuf) {
			buff.writeCollection(collection) { a, b -> b.toNetwork(a) }
		}

		override fun play(context: Supplier<NetworkEvent.Context>) {
			context.packetHandled = true

			if (NULLABLE_MINECRAFT_SERVER is IntegratedServer) {
				return
			}

			val builder = ImmutableMap.builder<ResourceLocation, AndroidResearchType>()

			for (v in collection) {
				builder.put(v.id, v)
			}

			researchMap = builder.build()

			for (value in researchMap.values) {
				value.validate()
			}

			context.enqueueWork {
				minecraft.player?.matteryPlayer?.reloadResearch()
			}
		}
	}

	fun readSyncPacket(buff: FriendlyByteBuf): SyncPacket {
		return SyncPacket(buff.readCollection({ LinkedList() }, AndroidResearchType.Companion::fromNetwork))
	}
}
