package ru.dbotthepony.mc.otm.client.render

import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonDeserializer
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import com.google.gson.JsonSerializer
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.client.screen.panels.DockProperty
import ru.dbotthepony.mc.otm.core.set
import java.lang.reflect.Type

@Suppress("unused")
data class SkinGrid(
	val texture: ResourceLocation,
	val width: Float,
	val height: Float,
	val rows: Int = 16,
	val columns: Int = 16,
) {
	var row = 0
	var column = 0

	val imageWidth get() = width * columns
	val imageHeight get() = height * rows

	val currentX: Float get() = column * width
	val currentY: Float get() = row * height

	fun skip(): SkinGrid {
		column++

		if (column >= columns) {
			jump()
		}

		return this
	}

	fun jump(): SkinGrid {
		row++
		column = 0
		return this
	}

	fun next(): SkinElement {
		val element = SkinElement(texture, currentX, currentY, width, height, imageWidth, imageHeight)
		skip()
		return element
	}

	fun reset(): SkinGrid {
		row = 0
		column = 0
		return this
	}

	fun seek(row: Int, column: Int): SkinGrid {
		this.row = row
		this.column = column
		return this
	}

	operator fun get(column: Int, row: Int) = SkinElement(texture, column * width, row * height, width, height, imageWidth, imageHeight)
}

fun ResourceLocation.element(
	x: Float,
	y: Float,
	width: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, height, textureWidth, textureHeight)

fun ResourceLocation.pixel(
	x: Float,
	y: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, 1f, textureWidth, textureHeight)

fun ResourceLocation.hLine(
	x: Float,
	y: Float,
	width: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, 1f, textureWidth, textureHeight)

fun ResourceLocation.vLine(
	x: Float,
	y: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, height, textureWidth, textureHeight)

@Suppress("unused")
data class SkinElement @JvmOverloads constructor(
	val texture: ResourceLocation,
	val x: Float,
	val y: Float,
	val width: Float,
	val height: Float,
	val imageWidth: Float = 256f,
	val imageHeight: Float = 256f,
	val winding: UVWindingOrder = UVWindingOrder.NORMAL
) {
	fun toJson(): JsonObject {
		return JsonObject().also {
			it["texture"] = JsonPrimitive(texture.toString())
			it["x"] = JsonPrimitive(x)
			it["y"] = JsonPrimitive(y)
			it["width"] = JsonPrimitive(width)
			it["height"] = JsonPrimitive(height)
			it["imageWidth"] = JsonPrimitive(imageWidth)
			it["imageHeight"] = JsonPrimitive(imageHeight)
			it["winding"] = JsonPrimitive(winding.name)
		}
	}

	fun toNetwork(buff: FriendlyByteBuf) {
		buff.writeUtf(texture.toString())
		buff.writeFloat(x)
		buff.writeFloat(y)
		buff.writeFloat(width)
		buff.writeFloat(height)
		buff.writeFloat(imageWidth)
		buff.writeFloat(imageHeight)
		buff.writeEnum(winding)
	}

	companion object : TypeAdapter<SkinElement>(), JsonSerializer<SkinElement>, JsonDeserializer<SkinElement> {
		fun fromNetwork(buff: FriendlyByteBuf): SkinElement {
			val texture = ResourceLocation(buff.readUtf())
			val x = buff.readFloat()
			val y = buff.readFloat()
			val width = buff.readFloat()
			val height = buff.readFloat()
			val imageWidth = buff.readFloat()
			val imageHeight = buff.readFloat()
			val winding = buff.readEnum(UVWindingOrder::class.java)

			return SkinElement(texture, x, y, width, height, imageWidth, imageHeight, winding)
		}

		fun fromJson(value: JsonObject): SkinElement {
			val texture = value["texture"]?.asString ?: throw JsonSyntaxException("Missing texture element")
			val x = value["x"]?.asFloat ?: throw JsonSyntaxException("Missing x element")
			val y = value["y"]?.asFloat ?: throw JsonSyntaxException("Missing y element")
			val width = value["width"]?.asFloat ?: throw JsonSyntaxException("Missing width element")
			val height = value["height"]?.asFloat ?: throw JsonSyntaxException("Missing height element")
			val imageWidth = value["imageWidth"]?.asFloat ?: throw JsonSyntaxException("Missing imageWidth element")
			val imageHeight = value["imageHeight"]?.asFloat ?: throw JsonSyntaxException("Missing imageHeight element")
			val winding = value["winding"]?.asString ?: throw JsonSyntaxException("Missing winding element")

			return SkinElement(ResourceLocation(texture), x, y, width, height, imageWidth, imageHeight, UVWindingOrder.valueOf(winding))
		}

		fun fromJson(value: JsonElement): SkinElement {
			return fromJson(value as? JsonObject ?: throw JsonSyntaxException("Expected JsonObject, got ${value::class.qualifiedName}"))
		}

		override fun write(out: JsonWriter, value: SkinElement) {
			TypeAdapters.JSON_ELEMENT.write(out, value.toJson())
		}

		override fun read(reader: JsonReader): SkinElement {
			val read = TypeAdapters.JSON_ELEMENT.read(reader)
			return fromJson(read as? JsonObject ?: throw JsonSyntaxException("Expected JsonObject, got ${read::class.qualifiedName}"))
		}

		override fun serialize(src: SkinElement, typeOfSrc: Type, context: JsonSerializationContext): JsonElement {
			return src.toJson()
		}

		override fun deserialize(
			json: JsonElement,
			typeOfT: Type,
			context: JsonDeserializationContext
		): SkinElement {
			return fromJson(json as? JsonObject ?: throw JsonSyntaxException("Expected JsonObject, got ${json::class.qualifiedName}"))
		}
	}

	init {
		require(x >= 0f) { "Invalid x $x" }
		require(y >= 0f) { "Invalid y $y" }
		require(width > 0f) { "Invalid width $width" }
		require(height > 0f) { "Invalid height $height" }
		require(imageWidth > 0f) { "Invalid image width $imageWidth" }
		require(imageHeight > 0f) { "Invalid image height $imageHeight" }
	}

	/**
	 * See [ru.dbotthepony.mc.otm.client.render.clearDepth]
	 */
	fun clearDepth(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
	) = ru.dbotthepony.mc.otm.client.render.clearDepth(stack, x, y, width, height)

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRaw(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = render(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRawPartial(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding
	) = renderPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		winding: UVWindingOrder = this.winding
	) {
		render(stack, x, y, width = width, winding = winding)
	}

	@JvmOverloads
	fun renderWidth(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		winding: UVWindingOrder = this.winding,
	) = renderWidth(stack, x.toFloat(), y.toFloat(), width.toFloat(), winding)

	@JvmOverloads
	fun renderHieght(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		height: Float = this.height,
		winding: UVWindingOrder = this.winding,
	) {
		render(stack, x, y, height = height, winding = winding)
	}

	@JvmOverloads
	fun renderHieght(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = this.winding,
	) = renderHieght(stack, x.toFloat(), y.toFloat(), height.toFloat(), winding)

	private val u0 = this.x / imageWidth
	private val v0 = this.y / imageHeight
	private val u1 = (this.x + width) / imageWidth
	private val v1 = (this.y + height) / imageHeight

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) {
		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) = renderRaw(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = this.width,
		height: Float = this.height,
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) {
		val u1 = (this.x + width) / imageWidth
		val v1 = (this.y + height) / imageHeight

		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = this.width.toDouble(),
		height: Double = this.height.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.NORMAL
	) = renderRawPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)
}

data class StretchingRectangleElement(
	val topLeft: SkinElement,
	val topRight: SkinElement,
	val bottomLeft: SkinElement,
	val bottomRight: SkinElement,
	val left: SkinElement,
	val right: SkinElement,
	val top: SkinElement,
	val bottom: SkinElement,
	val middle: SkinElement,
	val padding: DockProperty = DockProperty.EMPTY
) {
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float,
		height: Float,
	) {
		val bgWidth = width - topLeft.width.coerceAtMost(bottomLeft.width) - topRight.width.coerceAtMost(bottomRight.width) - padding.left - padding.right
		val bgHeight = height - topLeft.height.coerceAtMost(bottomLeft.height) - topRight.height.coerceAtMost(bottomRight.height) - padding.top - padding.bottom

		if (bgWidth > 0 && bgHeight > 0) {
			middle.render(stack, x + topLeft.width + padding.left, y + topLeft.height + padding.top, bgWidth, bgHeight)
		}

		top.render(stack, x + topLeft.width, y, width = width - topLeft.width - topRight.width)
		bottom.render(stack, x + bottomLeft.width, y + height - bottom.height, width = width - bottomLeft.width - bottomRight.width)

		left.render(stack, x, y + topLeft.height, height = height - topLeft.height - bottomLeft.height)
		right.render(stack, x + width - right.width, y + topRight.height, height = height - topRight.height - bottomRight.height)

		topLeft.render(stack, x, y)
		topRight.render(stack, x + width - topRight.width, y)

		bottomLeft.render(stack, x, y + height - bottomLeft.height)
		bottomRight.render(stack, x + width - bottomRight.width, y + height - bottomLeft.height)
	}

	companion object {
		fun square(
			texture: ResourceLocation,
			x: Float,
			y: Float,
			cornerWidth: Float,
			cornerHeight: Float,
			width: Float,
			height: Float,
			sideThickness: Float,
			imageWidth: Float = 256f,
			imageHeight: Float = 256f,
		): StretchingRectangleElement {
			val innerThickness = sideThickness.coerceAtLeast(cornerWidth).coerceAtLeast(cornerHeight)
			val padding = sideThickness - innerThickness

			return StretchingRectangleElement(
				topLeft = texture.element(x, y, cornerWidth, cornerHeight, imageWidth, imageHeight),
				topRight = texture.element(x + width - cornerWidth, y, cornerWidth, cornerHeight, imageWidth, imageHeight),
				bottomLeft = texture.element(x, y + height - cornerHeight, cornerWidth, cornerHeight, imageWidth, imageHeight),
				bottomRight = texture.element(x + width - cornerWidth, y + height - cornerHeight, cornerWidth, cornerHeight, imageWidth, imageHeight),

				top = texture.element(x + cornerWidth, y, width - cornerWidth * 2f, sideThickness, imageWidth, imageHeight),
				bottom = texture.element(x + cornerWidth, y + height - sideThickness, width - cornerWidth * 2f, sideThickness, imageWidth, imageHeight),
				left = texture.element(x, y, sideThickness, height - cornerHeight * 2f, imageWidth, imageHeight),
				right = texture.element(x, y + width - sideThickness, sideThickness, height - cornerHeight * 2f, imageWidth, imageHeight),

				middle = texture.element(x + innerThickness, y + innerThickness, width - innerThickness * 2f, height - innerThickness * 2f, imageWidth, imageHeight),
				padding = DockProperty(padding, padding, padding, padding)
			)
		}
	}
}

